/*************************************************
* X.509 Certificate Store Source File            *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/x509stor.h>
#include <botan/pubkey.h>
#include <botan/look_pk.h>
#include <botan/oids.h>
#include <botan/conf.h>
#include <memory>

namespace Botan {

namespace {

/*************************************************
* Compare the value of unique ID fields          *
*************************************************/
bool compare_ids(const SecureVector<byte>& id1,
                 const SecureVector<byte>& id2)
   {
   if(!id1.size() || !id2.size())
      return true;
   return (id1 == id2);
   }

}

/*************************************************
* Verify a certificate's authenticity            *
*************************************************/
X509_Code X509_Store::validate_cert(const X509_Certificate& cert)
   {
   add_cert(cert);

   std::vector<u32bit> indexes;
   X509_Code chaining_result = construct_cert_chain(cert, indexes);
   if(chaining_result != VERIFIED)
      return chaining_result;

   const u64bit current_time = system_time();

   for(u32bit j = 0; j != indexes.size() - 1; j++)
      {
      const X509_Certificate& current_cert = certs[indexes[j]].cert;
      s32bit validity_check = current_cert.check_validity(current_time);
      if(validity_check < 0)      return CERT_NOT_YET_VALID;
      else if(validity_check > 0) return CERT_HAS_EXPIRED;

      X509_Code sig_check_result = check_sig(certs[indexes[j]],
                                             certs[indexes[j+1]]);
      if(sig_check_result != VERIFIED)
         return sig_check_result;
      }
   return VERIFIED;
   }

/*************************************************
* Find this certificate                          *
*************************************************/
u32bit X509_Store::find_cert(const X509_Certificate& cert) const
   {
   for(u32bit j = 0; j != certs.size(); j++)
      {
      const X509_Certificate& this_cert = certs[j].cert;
      if(this_cert == cert)
         return j;
      }
   return NO_CERT_FOUND;
   }

/*************************************************
* Find the parent of this certificate            *
*************************************************/
u32bit X509_Store::find_parent_of(const X509_Certificate& cert) const
   {
   X509_DN issuer_info(cert.issuer);

   for(u32bit j = 0; j != certs.size(); j++)
      {
      const X509_Certificate& this_cert = certs[j].cert;
      X509_DN subject_info(this_cert.subject);
      if(issuer_info == subject_info)
         {
         if(!compare_ids(this_cert.v2_subject_key_id, cert.v2_issuer_key_id))
            continue;
         if(!compare_ids(this_cert.v3_subject_key_id, cert.v3_issuer_key_id))
            continue;
         return j;
         }
      }
   return NO_CERT_FOUND;
   }

/*************************************************
* Construct a chain of certificate relationships *
*************************************************/
X509_Code X509_Store::construct_cert_chain(const X509_Certificate& end_cert,
                                           std::vector<u32bit>& indexes) const
   {
   indexes.push_back(find_cert(end_cert));

   u32bit parent = find_parent_of(end_cert);

   while(true)
      {
      if(parent == NO_CERT_FOUND)
         return ISSUER_NOT_FOUND;
      indexes.push_back(parent);

      if(certs[parent].is_verified())
         return certs[parent].verify_result();

      const X509_Certificate& parent_cert = certs[parent].cert;
      if(!parent_cert.is_ca_cert())
         return CERT_NOT_FOR_CA;

      if(certs[parent].is_trusted())
         break;
      if(parent_cert.self_signed())
         return CANNOT_ESTABLISH_TRUST;

      if(parent_cert.path_len_constraint() < indexes.size() - 2)
         return CERT_CHAIN_TOO_LONG;

      parent = find_parent_of(parent_cert);
      }
   return VERIFIED;
   }

/*************************************************
* Check the CAs signature on a certificate       *
*************************************************/
X509_Code X509_Store::check_sig(const Cert_Info& cert,
                                const Cert_Info& ca_cert) const
   {
   if(cert.is_verified())
      return cert.verify_result();

   const X509_Certificate& this_cert = cert.cert;

   X509_Code verify_code = check_sig(this_cert, ca_cert);

   cert.set_result(verify_code);

   return verify_code;
   }

/*************************************************
* Check the CAs signature on a certificate       *
*************************************************/
X509_Code X509_Store::check_sig(const X509_Certificate& cert,
                                const Cert_Info& ca_cert) const
   {
   std::auto_ptr<X509_PublicKey> pub_key(ca_cert.cert.subject_public_key());
   std::auto_ptr<PK_Verifier> verifier;

   try {
      std::vector<std::string> sig_info =
         split_on_slash(OIDS::lookup(cert.sig_algo.oid));

      if(sig_info.size() != 2 && sig_info.size() != 3)
         return SIGNATURE_ERROR;
      if(sig_info[0] != pub_key->algo_name())
         return SIGNATURE_ERROR;

      std::string padding = sig_info[1];
      Signature_Format format = IEEE_1363;

      if(sig_info.size() == 3)
         {
         if(sig_info[2] == "DER")
            format = DER_SEQUENCE;
         else
            throw Decoding_Error("Unknown encoding method " + sig_info[2]);
         }

      if(dynamic_cast<PK_Verifying_with_MR_Key*>(pub_key.get()))
         {
         PK_Verifying_with_MR_Key* sig_key =
            dynamic_cast<PK_Verifying_with_MR_Key*>(pub_key.get());
         verifier.reset(get_pk_verifier(*sig_key, padding, format));
         }
      else if(dynamic_cast<PK_Verifying_wo_MR_Key*>(pub_key.get()))
         {
         PK_Verifying_wo_MR_Key* sig_key =
            dynamic_cast<PK_Verifying_wo_MR_Key*>(pub_key.get());
         verifier.reset(get_pk_verifier(*sig_key, padding, format));
         }
      else
         return CA_CERT_CANNOT_SIGN;

      bool valid = verifier->verify_message(cert.tbs_certificate(),
                                            cert.signature());

      if(valid)
         return VERIFIED;
      else
         return SIGNATURE_ERROR;
   }
   catch(Decoding_Error)
      {
      return CERT_FORMAT_ERROR;
      }
   catch(...)
      {
      return UNKNOWN_X509_ERROR;
      }
   return UNKNOWN_X509_ERROR;
   }

/*************************************************
* Search for certificates in the store           *
*************************************************/
std::vector<X509_Certificate> X509_Store::find(const std::string& email,
                                               const std::string& name) const
   {
   std::vector<X509_Certificate> matching_certs;

   for(u32bit j = 0; j != certs.size(); j++)
      {
      const X509_Certificate& this_cert = certs[j].cert;
      const std::string common_name = this_cert.subject_info("Name");
      const std::string email_addr = this_cert.subject_info("RFC822");

      if((email == "" || email_addr.find(email) != std::string::npos) &&
         (name  == "" || common_name.find(name) != std::string::npos))
         matching_certs.push_back(this_cert);
      }
   return matching_certs;
   }

/*************************************************
* Add a certificate to the store                 *
*************************************************/
void X509_Store::add_cert(const X509_Certificate& cert, bool trusted)
   {
   if(trusted && !cert.self_signed())
      throw Invalid_Argument("X509_Store: Trusted certs must be self-signed");

   if(find_cert(cert) == NO_CERT_FOUND)
      {
      Cert_Info info(cert, trusted);
      certs.push_back(info);
      }
   else if(trusted)
      {
      for(u32bit j = 0; j != certs.size(); j++)
         {
         const X509_Certificate& this_cert = certs[j].cert;
         if(this_cert == cert)
            certs[j].trusted = trusted;
         }
      }
   }

/*************************************************
* Add one or more certificates to the store      *
*************************************************/
void X509_Store::do_add_certs(DataSource& source, X509_Encoding encoding,
                              bool trusted)
   {
   while(!source.end_of_data())
      {
      try {
         X509_Certificate cert(source, encoding);
         add_cert(cert, trusted);
      }
      catch(X509_Certificate::X509_Error) {}
      catch(Decoding_Error) {}
      catch(Invalid_Argument) {}
      }
   }

/*************************************************
* Add one or more certificates to the store      *
*************************************************/
void X509_Store::add_certs(DataSource& source, X509_Encoding encoding)
   {
   do_add_certs(source, encoding, false);
   }

/*************************************************
* Add one or more certificates to the store      *
*************************************************/
void X509_Store::add_trusted_certs(DataSource& source, X509_Encoding encoding)
   {
   do_add_certs(source, encoding, true);
   }

/*************************************************
* PEM encode the set of certificates             *
*************************************************/
std::string X509_Store::PEM_encode() const
   {
   std::string cert_store;
   for(u32bit j = 0; j != certs.size(); j++)
      cert_store += certs[j].cert.PEM_encode();
   return cert_store;
   }

/*************************************************
* Create a Cert_Info structure                   *
*************************************************/
X509_Store::Cert_Info::Cert_Info(const X509_Certificate& c,
                                 bool t) : cert(c), trusted(t)
   {
   checked = false;
   result = UNKNOWN_X509_ERROR;
   last_checked = 0;
   }

/*************************************************
* Return the verification results                *
*************************************************/
X509_Code X509_Store::Cert_Info::verify_result() const
   {
   if(!checked)
      throw Invalid_State("Cert_Info::verify_result() called; not checked");
   return result;
   }

/*************************************************
* Set the verification results                   *
*************************************************/
void X509_Store::Cert_Info::set_result(X509_Code code) const
   {
   result = code;
   last_checked = system_time();
   checked = true;
   }

/*************************************************
* Check if this certificate can be trusted       *
*************************************************/
bool X509_Store::Cert_Info::is_trusted() const
   {
   return trusted;
   }

/*************************************************
* Check if this certificate has been verified    *
*************************************************/
bool X509_Store::Cert_Info::is_verified() const
   {
   if(!checked)
      return false;
   if(result != VERIFIED)
      return true;

   const u32bit CACHE_TIME = Config::get_time("x509/cache_verify_results");
   const u64bit current_time = system_time();

   if(current_time > last_checked + CACHE_TIME)
      checked = false;

   return checked;
   }

}
